import os

from .screen import Screen
from .defs import KEYMAP as _KEYMAP


# Standard widget result actions (as return from .loop())
ACTION_OK = 1000
ACTION_CANCEL = 1001
ACTION_NEXT = 1002
ACTION_PREV = 1003

class Widget(Screen):

    def __init__(self):
        self.kbuf = b""
        self.signals = {}

    def set_xy(self, x, y):
        self.x = x
        self.y = y

    def inside(self, x, y):
        return self.y <= y < self.y + self.h and self.x <= x < self.x + self.w

    def signal(self, sig):
        if sig in self.signals:
            self.signals[sig](self)

    def on(self, sig, handler):
        self.signals[sig] = handler

    @staticmethod
    def longest(items):
        if not items:
            return 0
        return max((len(t) for t in items))

    def set_cursor(self):
        # By default, a widget doesn't use text cursor, so disables it
        self.cursor(False)

    def get_input(self):
        if self.kbuf:
            key = self.kbuf[0:1]
            self.kbuf = self.kbuf[1:]
        else:
            key = os.read(0, 32)
            if key[0] != 0x1b:
                key = key.decode()
                self.kbuf = key[1:].encode()
                key = key[0:1].encode()
        key = _KEYMAP.get(key, key)

        if isinstance(key, bytes) and key.startswith(b"\x1b[M") and len(key) == 6:
            row = key[5] - 33
            col = key[4] - 33
            return [col, row]

        return key

    def handle_input(self, inp):
        if isinstance(inp, list):
            res = self.handle_mouse(inp[0], inp[1])
        else:
            res = self.handle_key(inp)
        return res

    def loop(self):
        self.redraw()
        while True:
            key = self.get_input()
            res = self.handle_input(key)

            if res is not None and res is not True:
                return res


class FocusableWidget(Widget):
    # If set to non-False, pressing Enter on this widget finishes
    # dialog, with Dialog.loop() return value being this value.
    finish_dialog = False


class EditableWidget(FocusableWidget):

    def get(self):
        raise NotImplementedError


class ChoiceWidget(EditableWidget):

    def __init__(self, choice):
        super().__init__()
        self.choice = choice

    def get(self):
        return self.choice


# Widget with few internal selectable items
class ItemSelWidget(ChoiceWidget):

    def __init__(self, items):
        super().__init__(0)
        self.items = items

    def move_sel(self, direction):
        self.choice = (self.choice + direction) % len(self.items)
        self.redraw()
        self.signal("changed")



# Colors

C_BLACK    = 0
C_RED      = 1
C_GREEN    = 2
C_YELLOW   = 3
C_BLUE     = 4
C_MAGENTA  = 5
C_CYAN     = 6
C_WHITE    = 7
ATTR_INTENSITY = 8
C_GRAY     = C_BLACK | ATTR_INTENSITY
C_B_RED      = C_RED | ATTR_INTENSITY
C_B_GREEN    = C_GREEN | ATTR_INTENSITY
C_B_YELLOW   = C_YELLOW | ATTR_INTENSITY
C_B_BLUE     = C_BLUE | ATTR_INTENSITY
C_B_MAGENTA  = C_MAGENTA | ATTR_INTENSITY
C_B_CYAN     = C_CYAN | ATTR_INTENSITY
C_B_WHITE    = C_WHITE | ATTR_INTENSITY

def C_PAIR(fg, bg):
    return (bg << 4) + fg

# Keys

KEY_UP = 1
KEY_DOWN = 2
KEY_LEFT = 3
KEY_RIGHT = 4
KEY_HOME = 5
KEY_END = 6
KEY_PGUP = 7
KEY_PGDN = 8
KEY_QUIT = 9
KEY_ENTER = 10
KEY_BACKSPACE = 11
KEY_DELETE = 12
KEY_TAB = b"\t"
KEY_SHIFT_TAB = b"\x1b[Z"
KEY_ESC = 20
KEY_F1 = 30
KEY_F2 = 31
KEY_F3 = 32
KEY_F4 = 33
KEY_F5 = b'\x1b[15~'
KEY_F6 = b'\x1b[17~'
KEY_F7 = b'\x1b[18~'
KEY_F8 = b'\x1b[19~'
KEY_F9 = b'\x1b[20~'
KEY_F10 = b'\x1b[21~'

KEYMAP = {
b"\x1b[A": KEY_UP,
b"\x1b[B": KEY_DOWN,
b"\x1b[D": KEY_LEFT,
b"\x1b[C": KEY_RIGHT,
b"\x1bOH": KEY_HOME,
b"\x1bOF": KEY_END,
b"\x1b[1~": KEY_HOME,
b"\x1b[4~": KEY_END,
b"\x1b[5~": KEY_PGUP,
b"\x1b[6~": KEY_PGDN,
b"\x03": KEY_QUIT,
b"\r": KEY_ENTER,
b"\x7f": KEY_BACKSPACE,
b"\x1b[3~": KEY_DELETE,
b"\x1b": KEY_ESC,
b"\x1bOP": KEY_F1,
b"\x1bOQ": KEY_F2,
b"\x1bOR": KEY_F3,
b"\x1bOS": KEY_F4,
}

# Unicode symbols in UTF-8

# DOWNWARDS ARROW
DOWN_ARROW = b"\xe2\x86\x93"
# BLACK DOWN-POINTING TRIANGLE
DOWN_TRIANGLE = b"\xe2\x96\xbc"



from .screen import Screen
from .basewidget import *
from .defs import *


class WMenuBar(ItemSelWidget):

    def __init__(self, menu_struct):
        super().__init__(menu_struct)
        self.x = self.y = 0
        self.h = 1
        self.w = Screen.screen_size()[0]
        self.pulled_down = False
        self.focus = False
        self.permanent = False

    def redraw(self):
        if self.focus:
            self.cursor(False)
        self.goto(self.x, self.y)
        i = 0
        for name, pulldown in self.items:
            if self.focus and i == self.choice:
                self.attr_color(C_B_WHITE, C_BLACK)
            self.wr(b"  ")
            self.wr(name)
            self.wr(b"  ")
            self.attr_reset()
            i += 1

    def close(self):
        self.focus = False
        if self.permanent:
            self.redraw()
        self.screen_redraw(True)

    def get_item_x(self, item_no):
        x = self.x
        i = 0
        for name, pulldown in self.items:
            if i == item_no:
                break
            x += len(name) + 4
            i += 1
        return x

    def handle_key(self, key):
        # We need while to auto pull down submenus on left/right keys
        while 1:
            res = None
            action = False
            sel = self.items[self.choice][1]
            if key == KEY_ESC:
                self.close()
                return ACTION_CANCEL
            elif key == KEY_LEFT:
                if self.pulled_down:
                    self.screen_redraw()
                self.move_sel(-1)
            elif key == KEY_RIGHT:
                if self.pulled_down:
                    self.screen_redraw()
                self.move_sel(1)
            elif key == KEY_ENTER:
                self.pulled_down = True
                action = True
            elif key == KEY_DOWN and isinstance(sel, Widget):
                self.pulled_down = True
            else:
                return

            sel = self.items[self.choice][1]
            if isinstance(sel, Widget) and self.pulled_down:
                sel.set_xy(self.get_item_x(self.choice), self.y + 1)
                res = sel.loop()
                if res == ACTION_PREV:
                    key = KEY_LEFT
                    continue
                if res == ACTION_NEXT:
                    key = KEY_RIGHT
                    continue
                if res == ACTION_CANCEL:
                    self.pulled_down = False
                    self.screen_redraw()
                    self.redraw()
                    return

                self.close()
                return res
            elif action:
                self.close()
                return sel

            return

    def handle_mouse(self, x, y):
        # Works in absolute coordinates
        if y != self.y:
            return
        x -= self.x
        cur_x = 0
        found = False
        i = 0
        for name, pulldown in self.items:
            cur_x += len(name) + 4
            if x <= cur_x:
                found = True
                break
            i += 1
        if not found:
            return
        self.choice = i
        self.redraw()
        return self.handle_key(KEY_ENTER)


class WMenuBox(ItemSelWidget):

    def __init__(self, items):
        super().__init__(items)
        self.x = self.y = 0
        self.h = len(items) + 2
        w = 0
        for i in items:
            w = max(w, len(i[0]))
        self.w = w + 2

    def redraw(self):
        self.dialog_box(self.x, self.y, self.w, self.h)
        i = 0
        for item in self.items:
            self.goto(self.x + 1, self.y + i + 1)
            if i == self.choice:
                self.attr_color(C_B_WHITE, C_BLACK)
            self.wr_fixedw(item[0], self.w - 2)
            self.attr_reset()
            i += 1

    def handle_key(self, key):
        if key == KEY_ESC:
            return ACTION_CANCEL
        elif key == KEY_UP:
            self.move_sel(-1)
        elif key == KEY_DOWN:
            self.move_sel(1)
        elif key == KEY_LEFT:
            return ACTION_PREV
        elif key == KEY_RIGHT:
            return ACTION_NEXT
        elif key == KEY_ENTER:
            return self.items[self.choice][1]

    def handle_mouse(self, x, y):
        if not self.inside(x, y):
            return ACTION_CANCEL
        y -= self.y + 1
        if 0 <= y < self.h - 2:
            self.choice = y
            return self.items[self.choice][1]



import sys
from setuptools import setup


if sys.version_info < (3, 0):
    sys.stderr.write("Sorry, Python < 3.0 is not supported\n")
    sys.exit(1)


setup(name='picotui',
      version='1.1',
      description="""A simple text user interface (TUI) library.""",
      long_description=open('README.rst').read(),
      url='https://github.com/pfalcon/picotui',
      author='Paul Sokolovsky',
      author_email='pfalcon@users.sourceforge.net',
      license='MIT',
      packages=['picotui'])



from .widgets import *


def add_ok_cancel_buttons(d):
    if d.h == 0:
        d.autosize(0, 1)
    if d.w < 20:
        d.w = 20
    hw = d.w // 2
    off1 = (hw + 1) // 2 - 4
    off2 = (d.w - hw) + hw // 2 - 4
    b = WButton(8, "OK")
    d.add(off1, d.h - 1, b)
    b.finish_dialog = ACTION_OK

    b = WButton(8, "Cancel")
    d.add(off2, d.h - 1, b)
    b.finish_dialog = ACTION_CANCEL


class DTextEntry(Dialog):

    def __init__(self, entry_w, text, title=""):
        super().__init__(10, 5, title=title)
        self.entry = WTextEntry(entry_w, text)
        self.entry.finish_dialog = ACTION_OK
        self.add(1, 1, self.entry)

    def result(self):
        res = self.loop()
        if res == ACTION_OK:
            return self.entry.get()


class DMultiEntry(Dialog):

    def __init__(self, entry_w, entry_h, lines, title=""):
        super().__init__(10, 5, entry_w + 2, entry_h + 3, title=title)
        self.widget = WMultiEntry(entry_w, entry_h, lines)
        self.add(1, 1, self.widget)
        add_ok_cancel_buttons(self)

    def result(self):
        res = self.loop()
        if res == ACTION_CANCEL:
            return res
        return self.widget.content


class DConfirmation(Dialog):

    def __init__(self, lines, title=""):
        super().__init__(10, 5, title=title)
        if not isinstance(lines, list):
            lines = [lines]
        i = 1
        for l in lines:
            self.add(2, i, WLabel(l))
            i += 1
        self.autosize(extra_w=1, extra_h=1)
        add_ok_cancel_buttons(self)

    def result(self):
        return self.loop()
